"""
Module for cloudwatch resources.
"""
import datetime

from .base_control import ResourceLister
from .common import BaseChart, Common


class MetricViewer(BaseChart):
    """
    Metric display control. Displays the selected metric as a bar graph.
    """

    prefix = "metric_view"
    title = "Metrics"

    def title_info(self):
        return f"{self.metric_dimension['Value']} {self.metric_name}"

    def __init__(self, *args, caller, metric=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.metric_name = metric["name"]
        self.metric_namespace = metric["namespace"]
        self.metric_dimension = {
            "Name": caller.dimension[0],
            "Value": caller.dimension[1],
        }
        self.load_data()

    def load_data(self):
        """
        Loads time series into the bar graph based on the metric name, namespace and dimension provided.
        """
        api_kwargs = {
            "MetricDataQueries": (
                [
                    {
                        "Id": "metric",
                        "MetricStat": {
                            "Metric": {
                                "Namespace": self.metric_namespace,
                                "MetricName": self.metric_name,
                                "Dimensions": [self.metric_dimension],
                            },
                            "Period": 300,
                            "Stat": "Average",
                        },
                    }
                ],
            ),
            "StartTime": (datetime.datetime.now() - datetime.timedelta(hours=24),),
            "EndTime": (datetime.datetime.now(),),
        }
        call = Common.generic_api_call(
            "cloudwatch",
            "get_metric_data",
            api_kwargs,
            "Retrieve metrics",
            "Cloudwatch",
            subcategory="Metrics",
            resource=f"{self.metric_namespace}/{self.metric_name}",
        )
        if call["Success"]:
            data = call["Response"]
            for idx in range(len(data["MetricDataResults"][0]["Timestamps"])):
                timestamp = data["MetricDataResults"][0]["Timestamps"][idx]
                value = data["MetricDataResults"][0]["Values"][idx]
                self.add_datapoint(timestamp, value)


class MetricLister(ResourceLister):
    """
    Lister resource for Cloudwatch Metrics.
    """

    prefix = "metric_list"
    title = "Metrics"

    resource_type = "metric"
    main_provider = "cloudwatch"
    category = "Cloudwatch"
    subcategory = "Metrics"
    list_method = "list_metrics"
    item_path = ".Metrics"
    columns = {
        "namespace": {"path": ".Namespace", "size": 16, "weight": 0, "sort_weight": 0},
        "name": {"path": ".MetricName", "size": 64, "weight": 1, "sort_weight": 1},
    }
    open_command = MetricViewer.opener
    open_selection_arg = "metric"

    def title_info(self):
        return f"{self.dimension[0]}={self.dimension[1]}"

    def __init__(
        self, *args, dimension, metric_namespace=None, metric_name=None, **kwargs
    ):
        self.dimension = dimension
        self.list_kwargs = {}
        if metric_namespace is not None:
            self.list_kwargs["Namespace"] = metric_namespace
        if metric_name is not None:
            self.list_kwargs["MetricName"] = metric_name
        super().__init__(*args, **kwargs)



"""
Module which contains AWS context-related resources.

Contexts are keypairs able to access different AWS accounts.
"""
from botocore import exceptions as botoerror

from .base_control import DeleteResourceDialog, OpenableListControl
from .common import Common, SessionAwareDialog
from .termui.control import Border
from .termui.dialog import DialogFieldLabel, DialogFieldText
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes


class ImportContextDialog(SessionAwareDialog):
    """
    Dialog control for importing a context from the environment.

    Attributes
    ----------
    caller : awsc.termui.control.Control
        Parent control which controls this dialog.
    error_label : awsc.termui.dialog.DialogFieldLabel
        Error label for displaying validation errors.
    name_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the name of the context.
    access_key : str
        The access key being imported.
    secret_key : str
        The secret key being imported.
    access_key_field : awsc.termui.dialog.DialogFieldLabel
        Field for displaying the access key being imported.
    secret_key_field : awsc.termui.dialog.DialogFieldLabel
        Field for displaying that a secret key is being imported. It is masked.
    """

    def __init__(self, parent, alignment, dimensions, *args, caller=None, **kwargs):
        self.accepts_inputs = True
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "New Context",
            Common.color("modal_dialog_border_title"),
        )
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.caller = caller
        self.set_title_label("Enter AWS context details")
        self.name_field = DialogFieldText(
            "Name:", **Common.textfield_colors("context_add")
        )
        self.add_field(self.name_field)
        creds = Common.Session.service_provider.env_session().get_credentials()
        if creds is None or not creds.access_key or not creds.secret_key:
            Common.Session.set_message(
                "No valid credentials in environment.", Common.color("message_error")
            )
            self.close()
            return
        self.access_key = creds.access_key
        self.secret_key = creds.secret_key
        self.access_key_field = DialogFieldLabel(
            [
                ("Access key: ", Common.color("generic")),
                (creds.access_key, Common.color("highlight")),
            ],
            centered=False,
        )
        self.add_field(self.access_key_field)
        self.secret_key_field = DialogFieldLabel(
            [
                ("Secret key: ", Common.color("generic")),
                ("*" * len(creds.secret_key), Common.color("highlight")),
            ],
            centered=False,
        )
        self.add_field(self.secret_key_field)

    def input(self, key):
        if not self.accepts_inputs:
            return True
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        try:
            sts = Common.Session.service_provider.whoami(
                keys={"access": self.access_key, "secret": self.secret_key}
            )
        except botoerror.ClientError as error:
            self.error_label.text = "Key verification failed"
            Common.clienterror(
                error,
                "Key verification failed",
                "Core",
                subcategory="STS",
                resource=self.access_key,
                set_message=False,
                access_key=self.access_key,
                secret_key="<PRESENT>",
            )
            return

        self.accepts_inputs = False

        Common.Configuration.add_or_edit_context(
            self.name_field.text, sts["Account"], self.access_key, self.secret_key
        )
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


class AddContextDialog(SessionAwareDialog):
    """
    Dialog control for adding a new context.

    Attributes
    ----------
    caller : awsc.termui.control.Control
        Parent control which controls this dialog.
    error_label : awsc.termui.dialog.DialogFieldLabel
        Error label for displaying validation errors.
    name_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the name of the context.
    access_key_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the access key.
    secret_key_field : awsc.termui.dialog.DialogFieldText
        Textfield for entering the secret key.
    """

    def __init__(self, parent, alignment, dimensions, *args, caller=None, **kwargs):
        self.accepts_inputs = True
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "New Context",
            Common.color("modal_dialog_border_title"),
        )
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.set_title_label("Enter AWS context details")
        self.name_field = DialogFieldText(
            "Name:",
            label_min=16,
            color=Common.color(
                "context_add_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "context_add_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "context_add_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
        )
        self.add_field(self.name_field)
        self.access_key_field = DialogFieldText(
            "Access key:",
            label_min=16,
            color=Common.color(
                "context_add_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "context_add_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "context_add_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
        )
        self.add_field(self.access_key_field)
        self.secret_key_field = DialogFieldText(
            "Secret key:",
            label_min=16,
            color=Common.color(
                "context_add_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "context_add_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "context_add_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
            password=True,
        )
        self.add_field(self.secret_key_field)
        self.caller = caller

    def input(self, key):
        if not self.accepts_inputs:
            return True
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        if self.access_key_field.text == "":
            self.error_label.text = "Access key cannot be blank."
            return
        if self.secret_key_field.text == "":
            self.error_label.text = "Secret key cannot be blank."
            return
        try:
            sts = Common.Session.service_provider.whoami(
                keys={
                    "access": self.access_key_field.text,
                    "secret": self.secret_key_field.text,
                }
            )
        except botoerror.ClientError as error:
            self.error_label.text = "Key verification failed"
            Common.clienterror(
                error,
                "Key verification failed",
                "Core",
                subcategory="STS",
                resource=self.access_key_field.text,
                set_message=False,
                access_key=self.access_key_field.text,
                secret_key="<PRESENT>",
            )
            return

        self.accepts_inputs = False

        Common.Configuration.add_or_edit_context(
            self.name_field.text,
            sts["Account"],
            self.access_key_field.text,
            self.secret_key_field.text,
        )
        self.close()

    def close(self):
        if self.caller is not None:
            self.caller.reload_contexts()
        super().close()


class ContextList(OpenableListControl):
    """
    Lister control for contexts.
    """

    prefix = "context_list"
    title = "Contexts"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_column("account id", 12)
        self.add_column("default", 7)

        self.reload_contexts()

    def reload_contexts(self):
        """
        Refreshes the list of contexts from configuration.
        """
        self.entries = []
        for context, data in Common.Configuration["contexts"].items():
            self.add_entry(
                ListEntry(
                    context,
                    **{
                        "account id": data["account_id"],
                        "default": "✓"
                        if context == Common.Configuration["default_context"]
                        else " ",
                    }
                )
            )
        if 0 <= self.selected < len(self.entries):
            return
        try:
            self.selected = (
                Common.Configuration["contexts"]
                .keys()
                .index(Common.Configuration["default_context"])
            )
        except ValueError:
            self.selected = 0

    @OpenableListControl.Autohotkey("a", "Add new context")
    def add_new_context(self, _):
        """
        Hotkey callback for adding a new context.
        """
        AddContextDialog.opener(caller=self)

    @OpenableListControl.Autohotkey("i", "Import context")
    def import_context(self, _):
        """
        Hotkey callback for importing a context.
        """
        ImportContextDialog.opener(caller=self)

    @OpenableListControl.Autohotkey("d", "Set as default", True)
    def set_default_context(self, _):
        """
        Hotkey callback for setting the default context.
        """
        if self.selection is not None:
            Common.Configuration["default_context"] = self.selection.name
            Common.Configuration.write_config()
            self.reload_contexts()

    @OpenableListControl.Autohotkey("KEY_ENTER", "Select", True)
    def select_context(self, _):
        """
        Hotkey callback for setting the active context.
        """
        if self.selection is not None:
            Common.Session.context = self.selection.name

    @OpenableListControl.Autohotkey(ControlCodes.D, "Delete context", True)
    def delete_selected_context(self, _):
        """
        Hotkey callback for deleting the selected context.
        """
        if self.selection is not None:
            DeleteResourceDialog.opener(
                caller=self,
                resource_type="context",
                resource_identifier=self.selection["name"],
                callback=self.do_delete,
            )

    def do_delete(self):
        """
        Action callback for the context deletion dialog.
        """
        Common.Configuration.delete_context(self.selection["name"])
        self.reload_contexts()



"""
Module for the encrypted key store object.
"""
import getpass
import hashlib
import os
import sys

import yaml
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.padding import PKCS7


class Keystore:
    """
    Encrypted key storage for AWS credentials.

    Attributes
    ----------
    keylist_file : pathlib.Path
        Encrypted binary which stores the AWS credentials.
    nonce_file : pathlib.Path
        Nonce file for the encryption.
    nonce : bytes
        The nonce value.
    keylist_cipher : cryptography.hazmat.primitives.cipher.Cipher
        The cryptography cipher used to decrypt and encrypt the key storage.
    keys : dict
        A list of decrypted key pairs, keyed by account name.
    """

    def __init__(self, config):
        """
        Initializes a Keystore object.

        Parameters
        ----------
        config : awsc.config.config.Configuration
            The parent configuration object instance.
        """
        self.keylist_file = config.path / "keys"
        self.nonce_file = config.path / "nonce"
        self.keylist_cipher = None
        self.nonce = os.urandom(16)
        self.keys = {}

    def unlock(self):
        """
        Prompts the user to enter the password to access the key storage, and parses the keylist file.
        """
        password = getpass.getpass("Enter encryption phrase for key database: ")
        sha = hashlib.sha256(password.encode("ascii"))
        if not self.nonce_file.exists():
            with self.nonce_file.open("wb") as file:
                file.write(self.nonce)
        else:
            with self.nonce_file.open("rb") as file:
                self.nonce = file.read()
        self.keylist_cipher = Cipher(AES(sha.digest()), CBC(self.nonce))

        if self.keylist_file.exists():
            self.parse_keylist()

    def parse_keylist(self):
        """
        Attempts to parse the keylist file.

        Should be called through unlock(), as unlock() generates the cipher for this method.
        """
        try:
            with self.keylist_file.open("rb") as file:
                data = file.read()
            dec = self.keylist_cipher.decryptor()
            yaml_decoded = dec.update(data) + dec.finalize()
            unpadder = PKCS7(256).unpadder()
            yaml_decoded = unpadder.update(yaml_decoded) + unpadder.finalize()
            self.keys = yaml.safe_load(yaml_decoded.decode("UTF-8"))
        except ValueError:
            print("Incorrect password.")
            sys.exit(1)

    def write_keylist(self):
        """
        Writes the keylist to the keylist_file. Cipher must be loaded through unlock() before writing.
        """
        enc = self.keylist_cipher.encryptor()
        yaml_encoded = yaml.dump(self.keys).encode("UTF-8")
        padder = PKCS7(256).padder()
        padded = padder.update(yaml_encoded) + padder.finalize()
        data = enc.update(padded) + enc.finalize()
        with self.keylist_file.open("wb") as file:
            file.write(data)

    def __getitem__(self, item):
        """
        Returns the named keypair.

        Parameters
        ----------
        item : str
            Name of the keypair to fetch.

        Returns
        -------
        dict
            A dict containing the access and secret keys.
        """
        return self.keys[item]

    def set_key(self, name, access, secret):
        """
        Upserts a key into the key storage and writes to disk.

        Parameters
        ----------
        name : str
            The name of the keypair.
        access : str
            The access key of the keypair.
        secret : str
            The secret key of the keypair.
        """
        self.keys[name] = {"access": access, "secret": secret}
        self.write_keylist()

    def delete_key(self, name):
        """
        Deletes a key from the key storage and writes to disk.

        Parameters
        ----------
        name : str
            The name of the key to delete.
        """
        del self.keys[name]
        self.write_keylist()



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from src.awsc.version import VERSION as version

with open("README.md", "r", encoding="utf-8") as readme:
    ld = readme.read()

setup(
    name="awsc",
    url="https://github.com/baliame/awsc",
    version=version,
    description="AWS Commander",
    long_description=ld,
    long_description_content_type="text/markdown",
    author="baliame",
    author_email="akos.toth@cheppers.com",  # Optional
    classifiers=[  # Optional
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
    ],
    package_dir={"": "src"},  # Optional
    packages=find_packages(where="src"),  # Required
    python_requires=">=3.8, <4",
    install_requires=[
        "blessed>=1.17.12",
        "gnupg>=1.3.2",
        "cryptography>=3.1",
        "pyyaml>=5.3.1",
        "boto3>=1.16.33",
        "jq>=1.1.1",
        "pyperclip>=1.8.1",
        "numpy>=1.17.4",
    ],
    entry_points={  # Optional
        "console_scripts": [
            "awsc=awsc:main",
        ],
    },
)
